
/************************************************************************** 
 *
 * extract_rtv.c  Read/write files on a ReplayTV hard drive
 *
 *   Release history:
 *  v1.0    12/12/2000    Original release
 *  v2.0    12/15/2000    Added faster extraction by walking down the
 *                           directory tree
 *                        Added command line option to select partition number
 *                        Added ability to extract just a subdirectory
 *                        Added ability to overwrite an existing file
 *  v3.0    12/16/2000    Changed reads/writes toF always be multiples of 512
 *                        for WinNT direct drive access compatibility
 *
 *  beta    2/16/2001     Autodetect ReplayTV drives, add support for 2-drive systems
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Many, many thanks to the guys on AVS Forum for all of the info on
 * the filesystem structures.
 * Especially wolfgang, videodude, vbcoder, toots, seanriddle
 * (in reverse alphabetical order)
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef WIN32  // Microsoft Visual C++

#include <io.h>
#include <winsock.h>
#include <direct.h>

#ifdef _DEBUG
#define WAIT { char buffer[5]; fprintf(stderr,"\nPress <ENTER> to exit\n"); fgets(buffer,sizeof(buffer),stdin); }
#else
#define WAIT
#endif

#define PACKED
#pragma pack(1)
typedef unsigned __int64 uint64;
#define FMT_LLU "%I64u"
#define FMT_LLU_12 "%12I64u"

#else  // Linux gcc

#define FMT_LLU "%llu"
#define FMT_LLU_12 "%12llu"
#include <unistd.h>
#include <linux/unistd.h>
#define WAIT
#define PACKED  __attribute__ ((packed))
#define O_BINARY 0
#define O_RANDOM 0
#pragma align 1
typedef unsigned long long uint64;

#endif

unsigned long SWAP32(unsigned long _s)
{
	return(   (((_s) >> 24) & 0x000000ff)
			| (((_s) >> 8)  & 0x0000ff00)
            | (((_s) << 8)  & 0x00ff0000)
			| (((_s) << 24) & 0xff000000) );
}

unsigned short SWAP16(unsigned short _s)
{
	return(   (((_s)&0xff00) >> 8)
			| (((_s)&0x00ff) << 8) );
}

uint64 SWAP64(uint64 net)
{
	uint64 res;
	unsigned long a,b;

	a=(unsigned long)(net>>32);
	b=(unsigned long)(net & 0xffffffffUL);
	a=SWAP32(a);
	b=SWAP32(b);
	res=(((uint64)b) << 32) + a;
	return res;
}

#include <string.h>

#include "crctest.h"

/* number of possible entries in the directory */
#define NUM_DIR_ENTRIES		0x49
#define NUM_EXTENT_ENTRIES  0x22

#define XOR_LENGTH 0x14
#define CRC_OFFSET 0x18

#define PATH_LENGTH 1024

int g_fd_disk1;
int g_fd_disk2;

uint64 g_offset;
uint64 g_clustersize;
uint64 g_clusters;
uint64 g_drive2start;
unsigned char *g_FAT;

int g_extract_all;
int g_directory_listing;

/*****************************************************************************
*
* The structure of a partition table
*
****************************************************************************/

struct partition_d
{
	char ipl[0x1be] PACKED;
	struct {
		unsigned char active PACKED;
		unsigned char start_chs[3] PACKED;
		unsigned char type PACKED;
		unsigned char end_chs[3] PACKED;
		unsigned int sector PACKED;
		unsigned int length PACKED;
	} table[4];
	short sig PACKED;
};


/*****************************************************************************
*
* The structure of a file/dir entry on the disk
*
****************************************************************************/

struct direntry_d
{
	uint64 self PACKED;		/* pointer to self */
	unsigned int endptr PACKED;
	unsigned short crc PACKED;		/* dont know how to compute yet */
	unsigned short unknown3 PACKED;
	unsigned short type2 PACKED;		/* no idea */
	unsigned char type3 PACKED;
	unsigned char xor PACKED;		/* completes checksum over above */
	unsigned int unknown4 PACKED;
	uint64 parent_dir PACKED;
	uint64 sibling PACKED;	/* Number of clusters in partition, for superblock */
	/* Sibling pointer for directories */
	uint64 rootdir PACKED;	/* Cluster number of root directory, for superblock */
	unsigned char padding[0x68] PACKED;
	unsigned char name[0x100] PACKED;	/* entry name */
	uint64 size PACKED;
	uint64 unknown7 PACKED;
	uint64 unknown8 PACKED;
	uint64 unknown9 PACKED;
	union {
		uint64 direntry[NUM_DIR_ENTRIES] PACKED;
		struct {
			uint64 unknown10[5] PACKED;
			uint64 extents[NUM_EXTENT_ENTRIES*2] PACKED;
		} file;
	} table;
};

/*
 * The Volume sequence block format:
 */

struct volume_sector
{
    unsigned long int signature PACKED;	/* Volume sequence block signature */
    unsigned long int version PACKED;	/* Volume sequence block version (this is a guess) */
    uint64 sequence PACKED;	/* Sequence number within volume */
    unsigned long int drive_count PACKED;	/* Number of drives in the volume (this is a guess) */
    unsigned long int filler1 PACKED;	/* Zeros */
    uint64 start_byte PACKED;	/* Starting byte within volume for this drive */
    uint64 exclude_bytes PACKED;	/* Number of bytes to exclude (on beginning of drive) */
    uint64 drive_size PACKED;	/* Size of drive in bytes */
    unsigned long int magic_number_1 PACKED;	/* First magic number field */
    unsigned long int magic_number_2 PACKED;	/* Second magic number field */
    char filler3[0x1c8] PACKED;	/* Fill out the rest of the sector */
};

#ifdef WIN32

void rtv_lseek(int fd,uint64 pos)
{
	uint64 newpos=_lseeki64(fd, pos, SEEK_SET);
	if(newpos == (uint64)(-1)) {
		perror("File seek error");
		WAIT;
		exit(-1);
	} else if(newpos != pos) {
		fprintf(stderr,"File seek error: wanted " FMT_LLU " got " FMT_LLU "\n",pos,newpos);
		WAIT;
		exit(-1);
	}
}

#else
_syscall5(int,_llseek,uint,fd,unsigned long,offset_high,
		  unsigned long,offset_low,loff_t *,result,uint,whence);

void rtv_lseek(int fd,uint64 pos)
{
	unsigned long a=(unsigned long)(pos>>32);
	unsigned long b=(unsigned long)(pos & 0xffffffff);

	pos=_llseek(fd,a,b,&pos,SEEK_SET);
}
#endif


/*****************************************************************************
 * 
 * reads one cluster from disk
 *
 ****************************************************************************/

void read_cluster(unsigned char *buf, uint64 len, uint64 cluster)
{
	uint64 pos;
	int count;
	int fd=g_fd_disk1;
	
	if(cluster > g_clusters) {
		fprintf(stderr,"Tried to read cluster " FMT_LLU ", but the limit is " FMT_LLU "\n",cluster,g_clusters);
		WAIT;
		exit(-1);
	}
	
	/* seek the current cluster */
	pos=g_offset + cluster * g_clustersize;

	/* For multiple-drive systems, decide which drive to use */
	if((g_drive2start > 0) && (pos >= g_drive2start)) {
		/* This cluster lies on the second drive */
		pos -= g_drive2start - 0x400;
		fd=g_fd_disk2;
		//fprintf(stderr,"Reading cluster " FMT_LLU " on disk 2\n",pos);
	}

	rtv_lseek(fd,pos);
	/* read current cluster */
	while(len>0) {
		/* For Win32 support, always read a complete 512-byte sector */
		count = read(fd, buf, 512);
		if(count>0) {
			buf+=count;
			len-=count;
		} else {
			if(count==0)
				fprintf(stderr,"EOF reading cluster " FMT_LLU "\n",cluster);
			else
				perror("Cluster read error");
			WAIT;
			exit(-1);
		}
	}
}

/*****************************************************************************
 *
 *  Do a CRC calculation
 *
 *****************************************************************************/

static cm_t cm;

void init_crc(void)
{
    // initialize CRC instance
    cm.cm_width = 16;
    cm.cm_poly = 0x1021;
    cm.cm_init = 0L;
    cm.cm_refin = FALSE;
    cm.cm_refot = FALSE;
    cm.cm_xorot = 0L;

    cm_ini(&cm);
}

unsigned short get_crc(const unsigned char *buf, int len)
{
	cm_ini(&cm);

    while (len--)
    {
        cm_nxt(&cm, *buf++);
    }

    return (unsigned short) cm_crc(&cm);
}
/*****************************************************************************
 * 
 * Mark off a cluster as used in the in-memory FAT, and check for crosslinks
 *
 ****************************************************************************/

void FAT_checkout(unsigned int cluster,unsigned int len,const char *name)
{
	unsigned int i;
	if(g_FAT==NULL) return;

	for(i=0;i<len;i++,cluster++) {
		if(cluster >= g_clusters) {
			fprintf(stderr,"Invalid cluster in (%.100s): %u\n",name,cluster);
			exit(-1);
		}
		if(++g_FAT[cluster] > 1) {
			fprintf(stderr,"Crosslinked cluster in (%.100s): %u\n",name,cluster);
			if(g_FAT[cluster] >= 5) {
				fprintf(stderr,"Too many crosslinks.  Aborting.\n");
				WAIT;
				exit(-1);
			}
		}
	}
}

/*****************************************************************************
 * 
 * Validate the integrity of an inode structure
 *
 ****************************************************************************/
int validate_inode(struct direntry_d *inode,int report_errors)
{
	char *ptr;
	unsigned int i,x,endptr;
	uint64 x64;
	unsigned short crc;

	/* Validate the entry using the XOR and CRC checks */
	ptr=(char *)inode;
	for(i=0,x=0;i<XOR_LENGTH;i++) {
		x = x ^ *(ptr++);
	}
	if(x != 0) {
		if(report_errors)
			fprintf(stderr,"Inode has XOR mismatch (0x%02x)\n",x);
		return 0;
	}
	endptr=SWAP32(inode->endptr);
	if(((endptr&0xf00f) != 0x0008) || (endptr<0x01d8) || (endptr>0x03e8)) {
		if(report_errors)
			fprintf(stderr,"Inode has an invalid endptr: 0x%08x, start=" FMT_LLU ", length=" FMT_LLU "\n",
					SWAP32(inode->endptr),
					SWAP64(inode->table.file.extents[0]),
					SWAP64(inode->size));
		return 0;
	}
	crc=get_crc(((char *)inode)+CRC_OFFSET,endptr);
	if(crc != SWAP16(inode->crc)) {
		if(report_errors)
			fprintf(stderr,"Inode has CRC mismatch (0x%04x should be 0x%04x)\n",crc,SWAP16(inode->crc));
		return 0;
	}
	// Validate the extents table
	if(endptr<0x3e8) {
		// Convert the end-of-entry pointer into a count
		endptr=((endptr&0x0ff0)>>4) - 0x1d;
		for(i = 0,x64=0;i <= endptr ; i++) {
			x64 += SWAP64(inode->table.file.extents[2*i+1]);
		}
		if(x64 != (uint64)(-1)) {
			if(report_errors)
				fprintf(stderr,"Inode has an invalid extents table\n");
			return 0;
		}
	}
	return 1;
}


/*****************************************************************************
 *
 * Find the inode for a directory/file entry
 *
 ****************************************************************************/
uint64 find_inode(struct direntry_d *root, const char *remotename)
{
	char buffer[256],*slash,*name;
	struct direntry_d dir;
	struct direntry_d inode;
	uint64 cluster,match;
	int idx;
	char ch;

	// Start with the root directory
	memcpy(&dir,root,sizeof(dir));

	// Keep a local copy of the name so we can mess with it
	strncpy(buffer,remotename,sizeof(buffer));
	buffer[sizeof(buffer)-1]=0;  // Make sure it is NULL-terminated

	name=buffer;

	for(ch='/' ; ch != 0 ; name=slash) {
		// Find the next slash, or end of string
		for(slash=name ; *slash && (*slash != '/') ; slash++);
		ch=*slash;
		*slash++=0;
		match=0;
		// Scan the current directory for this entry
		for(idx = 0; !match && (idx < NUM_DIR_ENTRIES) ; idx++) {
			cluster=SWAP64(dir.table.direntry[idx]);
			while((cluster != 0) && (cluster != (uint64)(-1)))
			{
				read_cluster((void *)&inode, sizeof(inode), cluster);
				if(!validate_inode(&inode,0)) {
					cluster=0;
				} else if(strcmp(name,inode.name)==0) {
					match=cluster;
					cluster=0;
				} else {
					// Look for a sibling, and continue
					cluster=SWAP64(inode.sibling);
				}
			}
		}
		if(!match)
			return 0;     // Only got a partial match
		if(ch==0)
			return match; // Have a complete match
		if(SWAP32(inode.endptr) != 0x3e8)
			return 0;     // Partial match is not a directory

		// Not at the end yet, so recurse
		memcpy(&dir,&inode,sizeof(dir));
	}
	return 0;  // Actually it can never get this far
}
						

/*****************************************************************************
 *
 * Replace file on the hard drive
 *
 ****************************************************************************/
int replace_file(uint64 cluster,const char *name)
{
	int i,x,wcnt,count,length,extent_count;
	unsigned int endptr;
	unsigned short crc;
	uint64 start,sublength,clusters,allocated,pos,*extents;
	unsigned char buffer[512],*ptr;
	struct direntry_d inode;
	FILE *fp;
	
	read_cluster((void *)&inode, sizeof(inode), cluster);
	if(!validate_inode(&inode,1)) {
		fprintf(stderr,"Inode for destination file in invalid.  Aborting\n");
		return 0;
	}
	endptr=SWAP32(inode.endptr);
	if(endptr==0x03e8) {
		fprintf(stderr,"Destination is a directory.  Can't insert files (yet)\n");
		return 0;
	}
	extent_count=(endptr >> 4) - 0x1d;
	if((extent_count<0)||(extent_count>0x20)) {
		fprintf(stderr,"Invalid extent table length in the destination inode\n");
		return 0;
	}

	fp = fopen(name, "rb");
	if(fp == NULL) {
		fprintf(stderr,"Local file '%s' does not exist.\n",name);
		return 0;
	}

	// Get the file length
	fseek(fp,0,SEEK_END);
	length=ftell(fp);
	rewind(fp);

	// Get the currently allocated space
	allocated=((uint64)(-1) - SWAP64(inode.table.file.extents[extent_count*2+1]))*g_clustersize;
	
	if(allocated < length) {
		fprintf(stderr,"Not enough allocated space in destination inode.\n  (Can't allocate new space yet)\n");
		return 0;
	}

	/* Change the length in the inode entry */
	inode.size=SWAP64(length);

	/* Compute the checksums for the updated inode */
	crc=get_crc(((char *)&inode+CRC_OFFSET),endptr);
	inode.crc=SWAP16(crc);
	ptr=(char *)&inode;
	for(i=0,x=0;i<(XOR_LENGTH-1);i++) {
		x = x ^ *(ptr++);
	}
	inode.xor=x;
	
	/* Validate that the new inode is ok */
	if(!validate_inode(&inode,1)) {
		fprintf(stderr,"Validation of the updated inode failed.  Aborting");
		return 0;
	}

	extents=inode.table.file.extents;
	while(length>0) {
		// Go through the cluster table until we get the full length
		start=SWAP64(*extents++);
		clusters=SWAP64(*extents++);
		
		if((start==(uint64)(-1)) || (clusters==(uint64)(-1))) {
			fprintf(stderr,"Error while writing to ReplayTV image:\n Found (-1)'s in the extent table\n",name);
			fclose(fp);
			return 0;
		}
		if(extent_count-- == 0) {
			fprintf(stderr,"Error while writing to ReplayTV image:\n Reached end of extent table\n",name);
			fclose(fp);
			return 0;
		}
		
		/* seek the current cluster */
		pos=g_offset + start * g_clustersize;
		rtv_lseek(g_fd_disk1,pos);
		
		sublength=clusters*g_clustersize;
		if(sublength>length) sublength=length;
		while(sublength>0) {
			count = fread(buffer,1,(sublength>sizeof(buffer)) ? sizeof(buffer) : (unsigned long)sublength,fp);
			if(count>0) {
				/* Fill incomplete sectors with zeros */
				for(i=count;i<sizeof(buffer);i++)
					buffer[i]=0;
				ptr=buffer;
				/* For Win32, always write a complete 512-byte sector */
				i=sizeof(buffer);
				do {
					wcnt=write(g_fd_disk1, ptr, i);
					i-=wcnt;
					ptr+=wcnt;
				} while((wcnt>0) && (i>0));
				if(i != 0) {
					if(wcnt<0)
						perror("Writing to ReplayTV partition");
					else
						fprintf(stderr,"Error writing to ReplayTV partition.  File corrupted!!!\n");
					fclose(fp);
					return 0;
				}
				length-=count;
				sublength-=count;
			} else {
				if(count==0)
					fprintf(stderr,"EOF reading file %s\n",name);
				else
					perror("File read error");
				fclose(fp);
				return 0;
			}
		}
	}
	fclose(fp);

	/* seek to the inode cluster */
	pos=g_offset + cluster * g_clustersize;
	rtv_lseek(g_fd_disk1,pos);
	/* Write the new inode to the ReplayTV disk */
	i=sizeof(inode);
	ptr=(char *)&inode;
	do {
		wcnt=write(g_fd_disk1, ptr, i);
		i-=wcnt;
		ptr+=wcnt;
	} while((wcnt>0) && (i>0));
	if(i != 0) {
		if(wcnt<0)
			perror("Writing to ReplayTV partition");
		else
			fprintf(stderr,"Error writing inode to ReplayTV partition.  File corrupted!!!\n");
		return 0;
	}
	return 1;
}

/*****************************************************************************
 * 
 * Copy file to local drive
 *
 ****************************************************************************/
void extract_file(uint64 *extents,unsigned int extent_count,uint64 length,const char *name)
{
	int count,fd,written;
	uint64 start,sublength,clusters,pos;
	unsigned char buffer[512];
	FILE *fp=NULL;
	
	fp = fopen(name, "rb");
	if(fp != NULL) {
		fclose(fp);
		fp=NULL;
		if(g_extract_all) {
			fprintf(stderr,"File '%s' already exists.  Skipping.\n",name);
			return;
		} else {
			printf("File '%s' exists.  Overwrite? ",name);
			fgets(buffer,sizeof(buffer),stdin);
			if(tolower(buffer[0])=='y')
				fp=fopen(name, "wb");
		}
	} else {
		fp=fopen(name, "wb");
	}
		
	if (fp == NULL) {
		fprintf(stderr,"Error writing file %s\n",name);
		WAIT;
		exit(-1);
	}
	
	while(length>0) {
		// Go through the cluster table until we get the full length
		start=SWAP64(*extents++);
		clusters=SWAP64(*extents++);
		
		if((start==(uint64)(-1)) || (clusters==(uint64)(-1))) {
			fprintf(stderr,"Error: Found (-1)'s in the extent table for %s\n",name);
			fclose(fp);
			return;
		}
		if(extent_count-- == 0) {
			fprintf(stderr,"Error: Reached end of extent table for %s\n",name);
			fclose(fp);
			return;
		}
		
		if((start+clusters-1) > g_clusters) {
			fprintf(stderr,"Tried to read clusters " FMT_LLU " to " FMT_LLU ", but the limit is " FMT_LLU "\n",
				start,start+clusters-1,g_clusters);
			WAIT;
			exit(-1);
		}
		
		/* seek the current cluster */
		pos=g_offset + start * g_clustersize;
		fd=g_fd_disk1;

		/* For multiple-drive systems, decide which drive to use */
		if((g_drive2start > 0) && (pos >= g_drive2start)) {
			/* This cluster lies on the second drive */
			pos -= g_drive2start - 0x400;
			fd=g_fd_disk2;
			//fprintf(stderr,"Reading cluster " FMT_LLU " on disk 2\n",pos);
		}
		rtv_lseek(fd,pos);

		sublength=clusters*g_clustersize;
		if(sublength>length) sublength=length;
		while(sublength>0) {
			/* For Win32 support, always read a complete 512-byte sector */
			count = read(fd, buffer, sizeof(buffer));
			if(count>0) {
				/* Only write the partial last sector to the local file */
				if(count > sublength)
					count=(unsigned int)sublength;
				written=fwrite(buffer,1,count,fp);

				if(written != count) {
					fprintf(stderr,"Error writing file %s  (Disk full?)\n",name);
					WAIT;
					exit(-1);
				}
				length-=count;
				sublength-=count;

				pos+=count;
				if((fd != g_fd_disk2) && (g_drive2start > 0) && (pos >= g_drive2start)) {
					/* Transition from first drive to second */
					pos -= g_drive2start - 0x400;
					fd=g_fd_disk2;
					rtv_lseek(fd,pos);
					//fprintf(stderr,"Reading cluster " FMT_LLU " on disk 2\n",pos);
				}
			} else {
				if(count==0)
					fprintf(stderr,"EOF reading file %s\n",name);
				else
					perror("Disk image read error");
				WAIT;
				exit(-1);
			}
		}
	}
	fclose(fp);
}



/*****************************************************************************
 * 
 * Recurse through the directory structure located in cluster
 *
 ****************************************************************************/

void recurse_directory(struct direntry_d *dir,const char *base)
{
	uint64 idx,cluster;
	struct direntry_d inode,inode2;
	char name[PATH_LENGTH],*ptr,*ptr2;
	unsigned int i,extent_count,err;
	
	for(idx = 0; idx < NUM_DIR_ENTRIES ; idx++) {
		cluster=SWAP64(dir->table.direntry[idx]);
		while((cluster != 0) && (cluster != (uint64)(-1)))
		{
			FAT_checkout((unsigned int)cluster,2,base);
			read_cluster((void *)&inode, sizeof(inode), cluster);
			read_cluster((void *)&inode2, sizeof(inode2), cluster+1);

			err=0;
			/* Make sure the two inodes are exactly the same */
			ptr=(char *)&inode;
			ptr2=(char *)&inode2;
			for(i=0;i<sizeof(inode);i++) {
				if(*ptr != *ptr2) {
					break;
				}
			}
			/* Validate the inode */
			if(i != sizeof(inode)) {
				fprintf(stderr,"Inode duplicates don't match\n");
				err=1;
			} else if(!validate_inode(&inode,1)) {
				err=1;
			}

			/* Extract the file name */
			inode.name[255]=0;  // Make sure this is null-terminated
			strncpy(name,base,sizeof(name));
			name[sizeof(name)-2]=0;
			strcat(name,"/");
			strncat(name,inode.name,sizeof(name)-strlen(name));
			name[sizeof(name)-1]=0;
			/* Can't allow '|' (pipe) characters in the filename under Unix or Windows */
			/* Can't allow ':' (colon) characters in Windows */
			for(ptr=name ; *ptr!=0 ; ptr++) {
				if(*ptr=='|')
					*ptr='!';
				if(*ptr==':')
					*ptr='.';
			}

			if(err) {
				fprintf(stderr," at cluster " FMT_LLU " (%.100s)\n",cluster,name);
				cluster=0;  // Don't follow siblings of faulty inodes
			} else {
				/* Comment if we see an interesting feature */
				if((inode.unknown3 != 0) && (inode.unknown3 != 0xffff)) {
					fprintf(stderr,"Inode at cluster " FMT_LLU " (%.100s) has a unknown3: %u\n",
						cluster,name,SWAP16(inode.unknown3));
				}
				if((inode.unknown4 != 0) && (inode.unknown4 != 0xffffffff)) {
					fprintf(stderr,"Inode at cluster " FMT_LLU " (%.100s) has a unknown4: %u\n",
						cluster,name,SWAP32(inode.unknown4));
				}
				if((inode.unknown7 != 0) && (inode.unknown7 != (uint64)(-1))) {
					fprintf(stderr,"Inode at cluster " FMT_LLU " (%.100s) has a unknown7: " FMT_LLU "\n",
						cluster,name,SWAP64(inode.unknown7));
				}
				if((inode.unknown8 != 0) && (inode.unknown8 != (uint64)(-1))) {
					fprintf(stderr,"Inode at cluster " FMT_LLU " (%.100s) has a unknown8: " FMT_LLU "\n",
						cluster,name,SWAP64(inode.unknown8));
				}
				if((inode.unknown9 != 0) && (inode.unknown9 != (uint64)(-1))) {
					fprintf(stderr,"Inode at cluster " FMT_LLU " (%.100s) has a unknown9: " FMT_LLU "\n",
						cluster,name,SWAP64(inode.unknown9));
				}

				extent_count=SWAP32(inode.endptr);
				switch(extent_count) {
				case 0x03e8: // Directory
					// Only create directories if we are extracting all files
					if(g_extract_all) {
						fprintf(stderr,"%-76.76s\r",name);
#ifdef WIN32
						if(mkdir(name) != 0)  // Arghh!  Microsoft always has to be different!
#else
						if(mkdir(name,0755) != 0)
#endif
							fprintf(stderr,"Unable to create directory %s\n",name);
					}
					recurse_directory(&inode,name);
					break;
				case 0x01d8: // zero-length file
					if(g_directory_listing) {
						printf("0\t%s\n",name);
					} else {
						fprintf(stderr,"Scan: %-70.70s\r",name);
					}
					if(g_extract_all)
						extract_file(0,0,0,name);
					break;
				default: // File
					// Convert the end-of-entry pointer into a count
					extent_count=((extent_count&0x0ff0)>>4) - 0x1d;

					// Mark out the extents in the FAT structure
					for(i=0;i<extent_count;i++) {
						FAT_checkout((unsigned int)SWAP64(inode.table.file.extents[2*i]),
								(unsigned int)SWAP64(inode.table.file.extents[2*i+1]),
								name);
					}
					if(g_directory_listing) {
						printf(FMT_LLU_12 "  %s\n",SWAP64(inode.size),name);
					} else {
						fprintf(stderr,"Scan: %-70.70s\r",name);
					}
					if(g_extract_all)
						extract_file(inode.table.file.extents,extent_count,SWAP64(inode.size),name);
				}
				cluster=SWAP64(inode.sibling);
			}
		}
	}
}

void scan_for_unlinked_inodes(void)
{
	struct direntry_d inode;
	uint64 cluster;

	printf("Scanning partition for deleted files.\n\n");
	for(cluster=0;cluster<g_clusters;cluster++) {
		if(g_FAT[cluster] == 0) {
			read_cluster((void *)&inode, sizeof(inode), cluster);

			/* Validate the entry using the XOR and CRC checks */
			if(validate_inode(&inode,0)) {
				inode.name[100]=0; // Make sure it is null-terminated and not too long
				printf("Valid inode (0x%04x) found at cluster " FMT_LLU " (%s)\n",
					SWAP32(inode.endptr),cluster,inode.name);
			}
		}
	}
}

/* I don't know how they would ever use this much space,
 * but it appears that some ReplayChannel files can be 500kB */
#define MAX_DESCRIPTION_LENGTH 1048576

struct FiledArray {
	int createtime;
	int filename;
	int type;
	int quality;
	int guaranteed;
	int exists;
	int index;
	int unknown1[6];
	char channelname[16];
	char channellabel[32];
	char lineup[8];
	int unknown2[6];
	int recordtime;
	int unknown3[5];
	char description[228];
	int duration;
};

void print_program_descriptions(uint64 cluster)
{
	struct direntry_d dir;
	struct direntry_d inode;
	int idx,i,extent_count,records,count,offset,shift;
	uint64 length,sublength,start,clusters,*extents;
	char *buffer,*ptr;
	struct FiledArray *record;
	time_t t;
	buffer=(char *)malloc(MAX_DESCRIPTION_LENGTH);
	if(buffer==NULL) {
		fprintf(stderr,"Memory allocation error in print_program_descriptions()\n");
		return;
	}

	read_cluster((void *)&dir, sizeof(dir), cluster);
	if(!validate_inode(&dir,1) || (SWAP32(dir.endptr)!=0x03e8)) {
		fprintf(stderr,"Invalid inode entry for ReplayChannels\n");
	} else {
		/* Recurse through the directory structure */
		for(idx = 0; idx < NUM_DIR_ENTRIES ; idx++) {
			cluster=SWAP64(dir.table.direntry[idx]);
			while((cluster != 0) && (cluster != (uint64)(-1)))
			{
				read_cluster((void *)&inode, sizeof(inode), cluster);
				// Convert the end-of-entry pointer into a count
				extent_count=((SWAP32(inode.endptr)&0x0ff0)>>4) - 0x1d;
				length = SWAP64(inode.size);

				if((length<0) || (length>MAX_DESCRIPTION_LENGTH) || (extent_count>0x20)) {
					fprintf(stderr,"Invalid ReplayChannel file.  Skipping ...\n");
				} else if(extent_count>0) {
					extents=inode.table.file.extents;
					offset=0;
					while((length>0) && (extent_count>0)) {
						/* Go through the cluster table until we get the full length */
						start=SWAP64(*extents++);
						clusters=SWAP64(*extents++);

						/* First partition will always be on the first drive,
						 * so we don't have to worry about changing disks */
						rtv_lseek(g_fd_disk1,g_offset + start * g_clustersize);
						sublength=clusters * g_clustersize;
						if(sublength > (length-offset))
							sublength = (length-offset);
						while(sublength > 0) {
							/* For Win32 support, always read a complete 512-byte sector */
							count = read(g_fd_disk1, buffer+offset, 512);
							if(count>0) {
								offset+=count;
								sublength-=count;
							} else {
								if(count==0)
									fprintf(stderr,"EOF reading ReplayChannel files\n");
								else
									perror("Disk image read error");
								WAIT;
								exit(-1);
							}
						}
						extent_count--;
					}

					/* Parse the description info out of this file */
					records = SWAP32(*(int *)(buffer+8));
					for(i=0,offset=16;(i<records) && (offset<length);i++) {
						while((offset<length) && (*(int *)(buffer+offset)==0))
							offset+=4;
						if(offset<length) {
							record=(struct FiledArray *)(buffer+offset);
							if(record->exists != 0) {
								ptr=buffer+offset+156;
								shift=0;
								while((ptr < (buffer+offset+400)) && (*ptr < 32)) {
									ptr+=4;
									shift+=4;
								}
								printf("%d.mpg\n",SWAP32(record->filename));
								/* Show title */
								printf("    %s\n",ptr);
								/* Show description */
								for( ; (ptr < (buffer+offset+400)) && *ptr ; ptr++);
								ptr++;
								printf("    %s\n",ptr);
								printf("    Source: %s (%s)\n",record->channellabel,record->channelname);
								/*
								t=SWAP32(record->createtime);
								printf("    Created: %s",ctime(&t));
								*/
								t=SWAP32(record->recordtime);
								printf("    Recorded: %s",ctime(&t));
								printf("    %d minutes at ",(SWAP32(record->duration)+30)/60);
								switch(SWAP32(record->quality)) {
								case 0: printf("High Quality"); break;
								case 1: printf("Medium Quality"); break;
								case 2: printf("Low Quality"); break;
								default: printf("Quality %d",SWAP32(record->quality));
								}
								switch(SWAP32(record->type)) {
								case 1: printf(" - Single record show\n\n"); break;
								case 2: printf(" - Zone show\n\n"); break;
								case 3: printf(" - Repeating show\n\n"); break;
								default: printf(" - Show type %d\n\n",SWAP32(record->type));
								}
								/*
								printf("    Guaranteed: %d\n",SWAP32(record->guaranteed));
								printf("    Exists: %d\n",SWAP32(record->exists));
								printf("    shift: %d\n",shift);
								printf("    Unknown1: %d %d %d %d %d %d\n",
									SWAP32(record->unknown1[0]),
									SWAP32(record->unknown1[1]),
									SWAP32(record->unknown1[2]),
									SWAP32(record->unknown1[3]),
									SWAP32(record->unknown1[4]),
									SWAP32(record->unknown1[5]));
								printf("    Unknown2: %d %d %d %d %d %d\n",
									SWAP32(record->unknown2[0]),
									SWAP32(record->unknown2[1]),
									SWAP32(record->unknown2[2]),
									SWAP32(record->unknown2[3]),
									SWAP32(record->unknown2[4]),
									SWAP32(record->unknown2[5]));
								printf("    Unknown3: %d %d %d %d %d\n",
									SWAP32(record->unknown3[0]),
									SWAP32(record->unknown3[1]),
									SWAP32(record->unknown3[2]),
									SWAP32(record->unknown3[3]),
									SWAP32(record->unknown3[4]));
								*/
							}
							offset+=404;
						}
					}
				}
				cluster=SWAP64(inode.sibling);
			}
		}
	}
	free(buffer);
}

/*****************************************************************************
 *
 * MAIN
 *
 ****************************************************************************/

int main(int argc, char **argv)
{
	struct direntry_d root;
	struct partition_d part;
	struct volume_sector sector2;
	char buffer[512];
	char *device1=NULL,*device2=NULL,*rtv_name=NULL,*local_name=NULL;
	int i;
	uint64 sectors,rootdir,cluster;

	/*
	 * Global flags
	 *   g_extract_all           extract all files with directory structure
	 *   g_directory_listing     print full file listing to standard out
	 */
	int partitiononly=0;         // drive image is a partition (no partition table)
	int verify_integrity=0;      // verify the integrity of the filesystem
	int find_unlinked_inodes=0;  // check for any inodes in the unused areas
	int extract=0;               // extract files from the replaytv drive
	int overwrite=0;             // overwrite files on the replaytv drive
	int partition=-1;            // used to select partition number on command line
	int program_descriptions=0;  // parse the program description listings

	if(argc < 2) {
		printf("extract_rtv built on " __DATE__ " " __TIME__ "\n\n");
		printf("USAGE: %s <devicename> <options>\n", argv[0]);
		printf("  -l            directory listing\n");
		printf("  -d            show program descriptions\n");
		printf("  -e            extract all files\n");
		printf("  -e filename   extract single file or subdirectory\n");
		printf("  -p            device file is a partition\n");
		printf("  -pN           use partition N (N=1,2,3,4)\n");
		printf("  -u            look for unlinked (deleted) files\n");
		printf("  -v            verify drive integrity\n");
		printf("  -w filename localfile\n");
		printf("                overwrite filename with localfile\n\n");
		printf("If you omit <devicename>, extract_rtv will attempt to\n");
		printf("autodetect the ReplayTV device connected to your system.\n\n");
		return -1;
	}
	
	g_extract_all = 0;
	
	for(i=1;i<argc;i++) {
		if(argv[i][0] == '-') {
			if(argv[i][1] == 'l') {         // listing only
				g_directory_listing = 1;
			} else if(argv[i][1] == 'd') {  // program descriptions
				program_descriptions = 1;
			} else if(argv[i][1] == 'e') {  // Extract files
				extract = 1;
				if(((i+1)<argc) && (argv[i+1][0] != '-')) {
					i++;
					rtv_name=argv[i];
					if(((i+1)<argc) && (argv[i+1][0] != '-')) {
						i++;
						local_name=argv[i];
					} else {
						/* Use just the filename portion of the remote name as the local name */
						local_name=rtv_name+strlen(rtv_name)-1;
						while( (local_name>=rtv_name) && (*local_name!='/') && (*local_name!='\\') )
							local_name--;
						local_name++;
					}
				} else {
					g_extract_all=1;
				}
			} else if(argv[i][1] == 'p') {  // Use partition dump (no part. table)
				if(argv[i][2]==0)
					partitiononly = 1;
				else
					partition=argv[i][2]-'1';
			} else if(argv[i][1] == 'u') {  // Unlinked inode check
				find_unlinked_inodes = 1;
				verify_integrity = 1;
			} else if(argv[i][1] == 'v') {  // Verify drive integrity
				verify_integrity = 1;
			} else if(argv[i][1] == 'w') {  // Overwrite file on the ReplayTV drive image
				overwrite = 1;
				if(((i+2)<argc) && (argv[i+1][0] != '-') && (argv[i+2][0] != '-')) {
					rtv_name=argv[i+1];
					local_name=argv[i+2];
					i+=2;
				} else {
					fprintf(stderr,"-w option requires both remote and local filenames\n");
					WAIT;
					return(-1);
				}
			} else {
				fprintf(stderr,"Ignored command line option: %s\n",argv[i]);
			}
		} else {
			if(device1==NULL)
				device1=argv[i];
			else if(device2==NULL)
				device2=argv[i];
			else
				fprintf(stderr,"Ignored command line option: %s\n",argv[i]);
		}
	}

	if(    (extract && overwrite)
		|| (program_descriptions && overwrite))
	{
		fprintf(stderr, "You can't extract and overwrite in the same operation!\n");
		WAIT;
		return -1;
	}

	if(program_descriptions) {
		/* Force partition 0 for program descriptions */
		if(partition>0) {
			fprintf(stderr,"Forcing to first partition to read program descriptions.\n");
		}
		partition=0;
	}

	if(device1==NULL) {
		/* Try to autodetect the device for ReplayTV drives */
		int fd;
		char *tmp=malloc(32);
#ifdef WIN32
		for(i=0;i<16;i++) {
			sprintf(tmp,"\\\\.\\PhysicalDrive%d",i);
#else /* Linux */
		for(i=0;i<4;i++) {
			sprintf(tmp,"/dev/hd%c",'a'+i);
#endif
			/* Read the partition table */
			memset(&part,0xFF,sizeof(part));
			memset(&sector2,0,sizeof(sector2));
			fd=open(tmp, O_RDONLY|O_BINARY);
			if(fd>0) {
				read(fd,&part,sizeof(part));
				read(fd,&sector2,sizeof(sector2));
				if((part.table[0].type==0x4D) && (part.table[1].type==0x4D)) {
					fprintf(stderr, "ReplayTV drive detected: %s\n",tmp);
					if(device1==NULL) {
						device1=tmp;
						tmp=malloc(32);  /* Leak a little memory */
					} else {
						fprintf(stderr, "Autodetect failed because multiple ReplayTV drives were found.\n");
						WAIT;
						return(-1);
					}
				}
				if(   (SWAP32(sector2.signature) == 0xD69EF5A0)
					&& (SWAP32(sector2.drive_count) == 2)
					&& (SWAP64(sector2.sequence) == 1)) {
					fprintf(stderr, "Secondary ReplayTV drive detected: %s\n",tmp);
					if(device2==NULL) {
						device2=tmp;
						tmp=malloc(32);  /* Leak a little memory */
					} else {
						fprintf(stderr, "Autodetect failed because multiple Secondary ReplayTV drives were found.\n");
						WAIT;
						return(-1);
					}
				}
				close(fd);
			}
		}
		free(tmp);
	}

	if(device1==NULL) {
		fprintf(stderr, "Didn't detect any ReplayTV drives connected to the system\n");
		WAIT;
		return -1;
	}

	g_fd_disk1 = open(device1, O_RDONLY|O_BINARY|O_RANDOM);
	if (g_fd_disk1 < 0) {
		perror("Unable to open drive image file");
		WAIT;
		return(-1);
	}
		
	g_offset=0;
	g_clustersize=512;
	g_clusters=2;
	g_FAT=NULL;
	g_drive2start=0;
	g_fd_disk2=0;

	/* Initialize the CRC code */
	init_crc();

	if(!partitiononly) {
		/* Read the partition table */
		read_cluster((void *)&part,sizeof(part),0);
		
		/* Print the partition table */
		if((partition<0) || (partition>3)) {
			fprintf(stderr,"Partition   Type        Start      Length\n");
			for(i=0;i<4;i++) {
				fprintf(stderr,"%d           0x%02x   %10d  %10d\n",i+1,
					part.table[i].type,
					part.table[i].sector,
					part.table[i].length);
			}
			fprintf(stderr,"\nSelect partition: ");
			fgets(buffer,10,stdin);
			partition=atoi(buffer)-1;
		}
		if((partition<0) || (partition>3) || (part.table[partition].type != 0x4D)) {
			fprintf(stderr,"Invalid partition selection.\n");
			WAIT;
			exit(-1);
		}
		/* Get the volume sequencing information for 2-drive systems */
		memset(&sector2,0,sizeof(sector2));
		read_cluster((char *)&sector2,sizeof(sector2),1);

		/* Compute the beginning of the partition */
		g_offset=part.table[partition].sector * 0x200;
		/* Compute the maximum number of sectors in the partition */
		sectors=part.table[partition].length;

		if(SWAP32(sector2.signature)==0xD69EF5A0) {
			if(device2!=NULL) {
				/* Extract drive1 size from volume signature block */
				g_drive2start=SWAP64(sector2.drive_size);

				/* Open the second drive */
				g_fd_disk2 = open(device2, O_RDONLY|O_BINARY|O_RANDOM);
				if (g_fd_disk2 < 0) {
					perror("Unable to open secondary drive image file");
					WAIT;
					return(-1);
				}
				/* Verify the second drive signature */
				memset(&part,0,sizeof(part));
				memset(&sector2,0,sizeof(sector2));
				read(g_fd_disk2,&part,sizeof(part));
				read(g_fd_disk2,&sector2,sizeof(sector2));
				/*
				if(    (part.table[0].type!=0)
					|| (part.table[1].type!=0)
					|| (part.table[2].type!=0)
					|| (part.table[3].type!=0)
					|| (part.sig != 0xAA55)) {
					fprintf(stderr,"WARNING: Partition table for second drive has non-ReplayTV information.\n");
				}
				*/
				if(SWAP32(sector2.signature)!=0xD69EF5A0) {
					fprintf(stderr,"WARNING: Second drive does not contain a ReplayTV signature!\n");
				}
			} else if(partition>0) {
				fprintf(stderr,"WARNING: ReplayTV drive appears to be part of a multi-drive system\n"
							   "         but only the first drive was found!\n"
							   "         Only files on the first drive can be extracted.\n");
			}
		}
		if(overwrite && (partition>0) && (g_fd_disk2>0)) {
			fprintf(stderr,"WARNING: Overwriting files on the second partition of multi-drive systems\n"
						   "         will only work if the file is entirely on the first drive!!!\n");
		}
	}
	
	if(overwrite) {
		/* Ask for confirmation */
		fprintf(stderr,"Overwriting a file on the ReplayTV drive is VERY RISKY!\n"
			"Are you sure you want to do this? ");
		fgets(buffer,sizeof(buffer),stdin);
		if(tolower(buffer[0]) != 'y') {
			fprintf(stderr,"Confirmation failed.  Aborting.\n");
			WAIT;
			return(-1);
		}

		//verify_integrity = 1;    // Force an integrity check
		g_directory_listing = 0; // Don't do a directory listing
	}

	/* Read the first sector of the partition */
	read_cluster(buffer,sizeof(buffer),0);
	
	/* Get the number of clusters and cluster size for the partition */
	g_clusters=SWAP64(*(uint64 *)(buffer+0x108));
	g_clustersize=SWAP32(*(unsigned int *)(buffer+0x114));
	
	if(verify_integrity) {
		/* Create the file allocation table */
		g_FAT=(unsigned char *)malloc((unsigned int)g_clusters*sizeof(unsigned char));
		if(g_FAT==NULL) {
			fprintf(stderr,"Memory allocation error while allocating FAT\n");
			WAIT;
			exit(-1);
		}
		for(i=0;i<g_clusters;i++)
			g_FAT[i]=0;
		/* Mark off the pre-defined FAT entries */
		FAT_checkout(0,3,"Superblock");
	}

	/* Make sure the clusters and sectors values are consistent */
	if(!partitiononly && (sectors < ((g_clusters*g_clustersize) >> 9))) {
		fprintf(stderr,"ERROR: Number of sectors (" FMT_LLU ") is less than clusters*clustersectors (" FMT_LLU ")\n",sectors,g_clusters*g_clustersize);
		WAIT;
		exit(-1);
	}
	
	/* Read the first superblock */
	read_cluster((void *)&root, sizeof(root), 1);
	
	/* Confirm that the length matches the value from the first sector */
	/* Note: the sibling entry in the 1st sector is the number of clusters */
	if(g_clusters != SWAP64(root.sibling)) {
		fprintf(stderr,"ERROR: 1st sector and superblock disagree on number of clusters\n");
		WAIT;
		exit(-1);
	}

	rootdir = SWAP64(root.rootdir);

	if(rootdir >= g_clusters) {
		fprintf(stderr,"Invalid root directory cluster: " FMT_LLU "\n",rootdir);
		WAIT;
		exit(-1);
	}
	/* Mark the root directory entry as taken in the FAT */
	FAT_checkout((unsigned int)rootdir,2,"Rootdir");

	/* Read the root directory entry */
	read_cluster((void *)&root, sizeof(root), rootdir);

	/* Start at the root dir, and recurse through */
	if(verify_integrity || g_directory_listing || g_extract_all ) {
		recurse_directory(&root,".");
		fprintf(stderr,"\n");
		if(verify_integrity)
			fprintf(stderr,"Filesystem integrity check complete.\n");
	}
	if(find_unlinked_inodes) {
		/* Go through the rest of the disk and look for valid inodes */
		fprintf(stderr,"Scanning for unlinked inodes.\n");
		scan_for_unlinked_inodes();
	}

	/* Overwrite a single file by walking the directory structure */
	if(overwrite) {
		/* Reopen the file read-write */
		close(g_fd_disk1);
		g_fd_disk1 = open(device1, O_RDWR|O_BINARY|O_RANDOM);
		if(g_fd_disk1<0) {
			perror("Unable to open drive image file in read/write mode");
			fprintf(stderr,"Do you have write permissions for %s ?\n",device1);
			WAIT;
			return(-1);
		}
		/* Strip the leading './' */
		while(*rtv_name && ((*rtv_name=='.')||(*rtv_name=='/'))) rtv_name++;
		cluster=find_inode(&root,rtv_name);
		if((cluster>0) && replace_file(cluster,local_name)) {
			fprintf(stderr,"Replaced file %s with %s\n",rtv_name,local_name);
		} else {
			fprintf(stderr,"File replacement failed\n");
		}
	}

	/* Done with the g_FAT array, so get rid of it
	 * Otherwise, using -v with -e for a subdirectory will
	 * report everything as crosslinked
	 */
	if(g_FAT != NULL) {
		free(g_FAT);
		g_FAT=NULL;
	}

	/* Extract a single file by walking the directory structure */
	if(extract && !g_extract_all) {
		/* Strip the leading './' */
		while(*rtv_name && ((*rtv_name=='.')||(*rtv_name=='/'))) rtv_name++;
		cluster=find_inode(&root,rtv_name);
		if(cluster>0) {
			read_cluster((void *)&root, sizeof(root), cluster);
			if(!validate_inode(&root,1)) {
				fprintf(stderr,"Invalid inode entry for %s\n",rtv_name);
			} else {
				i=SWAP32(root.endptr);
				if(i==0x03e8) {
					// Directory entry, extract its contents
					g_extract_all=1;
					fprintf(stderr,"Extracting directory %s\n",rtv_name);
					recurse_directory(&root,".");
				} else {
					i=(i >> 4) - 0x1d;
					if((i<0)||(i>0x20)) {
						fprintf(stderr,"Invalid extent table length for %s\n",rtv_name);
					} else {
						fprintf(stderr,"Extracting file %s\n",rtv_name);
						extract_file(root.table.file.extents,i,SWAP64(root.size),local_name);
					}
				}
			}
		} else {
			fprintf(stderr,"Couldn't find the file '%s'\n",rtv_name);
		}
	}

	/* Print the program descriptions for the stored MPEG files */
	if(program_descriptions) {
		cluster=find_inode(&root,"user/ReplayChannels/2/ReplayChannels");
		if(cluster>0)
			print_program_descriptions(cluster);
		else
			fprintf(stderr,"Couldn't find the ReplayChannels list\n");
	}

	close(g_fd_disk1);
	if(g_fd_disk2>0)
		close(g_fd_disk2);
	
	WAIT;
	return 0;
}
